open Utils
open Syntax_tree

type environment = {add_op : string -> int -> unit;
		    add_var : string -> unit;
		    add_rule : string -> string -> unit;
		    get_op_arity : string -> int;
		    is_op : string -> bool;
		    define : string -> Syntax_tree.base_expr -> unit;
		    is_defined : string -> bool;
		    get_def : string -> Syntax_tree.base_expr;
		    get_ops : unit -> (string * int) list;
		    get_vars :unit -> string list;
		    get_defs : unit -> (string * Syntax_tree.base_expr) list;
		    get_rules : unit -> (string * string) list;
		    set_current_dir : string -> unit;
		    get_current_dir : unit -> string;
		    push : (string * Syntax_tree.base_expr) list -> unit;
		    pop : string list -> unit;
		    defun : string -> string list -> Syntax_tree.insts -> unit;
		    apply_fun : string -> Syntax_tree.expr list -> Syntax_tree.inst;
		    is_fun : string -> bool;
		    get_funs : unit -> (string * (string list * Syntax_tree.insts)) list;
		    set_lines : (int * int) list -> unit;
		    get_line : unit -> int;
		    set_stream : char Stream.t -> unit;
		    set_stream_line : unit -> unit;
		    set_program_name : string -> unit;
		    get_program_name : unit -> string;
		   }

let get_new_environment() =
  let op_table = ref []
  and var_table = ref []
  and def_table = ref []
  and rule_table = ref []
  and fun_table = ref []
  and lines_table = ref []
  and line = ref 1
  and stream = ref [<>]
  and p_name = ref ""
  and current_dir = ref "." in
  let add_op op arity =
    if not(List.mem_assoc op !op_table) then
      op_table := (op,arity) :: !op_table
    else
      let c_arity = List.assoc op !op_table in
	if not(c_arity = arity || c_arity < 0 || arity < 0) then
	  failwith ("Operator " ^ op ^ " defined with arity " ^
		      (string_of_int c_arity) ^ ", trying to set arity " ^
		      (string_of_int arity))
	else op_table := (op,max c_arity arity) :: (List.remove_assoc op !op_table)
  and add_var var =
    if not(List.mem var !var_table) then
      var_table := var :: !var_table
    else ()
  and add_rule sb se =
    (let del_empty_seq s =
       Str.global_replace (Str.regexp "()") "" s in
       (let (sbn,sen) =
	  (del_empty_seq sb, del_empty_seq se) in
	  if (List.mem_assoc sbn !rule_table
	      && List.assoc sbn !rule_table = sen) then
	    ()
	  else
	    rule_table := (sbn,sen) :: !rule_table)
    )
  and get_op_arity op =
    List.assoc op !op_table
  and is_op op =
    List.mem_assoc op !op_table
  and define name value =
    def_table := (name,value) :: List.remove_assoc name !def_table
  and is_defined name =
    List.mem_assoc name !def_table
  and get_def name =
    try
      List.assoc name !def_table
    with _ -> failwith ("Variable $" ^ name ^ " is not defined")
  and get_ops () = List.rev !op_table
  and get_vars () = List.rev !var_table
  and get_defs () = List.rev !def_table
  and get_rules () = List.rev !rule_table
  and get_funs () = List.rev !fun_table
  and set_current_dir dir = current_dir := dir
  and get_current_dir () = !current_dir
  and push defs_vals =
    List.iter (fun x -> def_table := (fst x, snd x) :: !def_table) defs_vals
  and pop defs =
    List.iter (fun x -> def_table := List.remove_assoc x !def_table) defs
  and defun f prms insts =
    fun_table := (f,(prms,insts)) :: !fun_table
  and apply_fun f evprms =
    if not(List.mem_assoc f !fun_table) then
      failwith ("function $" ^ f ^ " is not defined");
    let elen = List.length evprms
    and (prms,insts) = List.assoc f !fun_table in
    let len = List.length prms in
      if (elen > len) then
	failwith ("Too much parameters applied to function " ^ f)
      else if (len > elen) then
	failwith ("Missing parameters for function " ^ f)
      else
	let prm_val_list = (List.combine prms evprms) in
	  LetIn(prm_val_list, insts)
  and is_fun f = List.mem_assoc f !fun_table
  and set_lines l = lines_table := l
  and get_line () = !line
  and set_stream s = stream := s
  and set_stream_line () =
    let c = Stream.count !stream in
    let i = ref 0
    and res = ref (1,1) in
      while (c > fst !res) do
	res := List.nth !lines_table !i;
	i := !i + 1
      done;
      line := if (!res = (1,1)) then 1 else (snd !res) - 1
  and set_program_name s = p_name := s
  and get_program_name () = !p_name
  in
    {add_op = add_op; add_var = add_var; add_rule = add_rule;
     get_op_arity = get_op_arity; is_op = is_op; define = define;
     is_defined = is_defined; get_def = get_def; get_ops = get_ops;
     get_vars = get_vars; get_defs = get_defs; get_rules = get_rules;
     set_current_dir = set_current_dir; get_current_dir = get_current_dir;
     push = push; pop = pop; defun = defun; apply_fun = apply_fun;
     is_fun = is_fun; get_funs = get_funs; set_lines = set_lines;
     get_line = get_line; set_stream = set_stream;
     set_stream_line = set_stream_line; set_program_name = set_program_name;
     get_program_name = get_program_name}

let merge_environments envref env =
  List.iter (fun x -> envref.add_op (fst x) (snd x)) (env.get_ops());
  List.iter (fun x -> envref.add_var x) (env.get_vars());
  List.iter (fun x -> envref.add_rule (fst x) (snd x)) (env.get_rules());
  List.iter (fun x -> envref.define (fst x) (snd x)) (env.get_defs());
  List.iter (fun x -> envref.defun (fst x) (fst(snd x)) (snd(snd x)))
    (env.get_funs())

let print_debug env =
  print_list " Ops " (fst (List.split (env.get_ops())));
  print_list " Vars " (env.get_vars());
  print_assoc_list " Defs "
    (List.map (fun (x,y) -> (x,string_of_base_expr y))
       (env.get_defs()))

let parsed_programs = ref []

let get_parsed_programs () =
  !parsed_programs

let set_parsed_program p env =
  parsed_programs := (p,env) :: !parsed_programs

let debug_mode = ref false
