(** Provides environments to store and modify data associated with program parsing. *)

(** Record containing functions to store, get and modify underlying data structures*)
type environment = {
  add_op : string -> int -> unit;
  (** [add_op op n] adds an operator [op] of arity [n] in the environment *)
  add_var : string -> unit;
  (** [add_var var] adds a variable [var] in the environment *)
  add_rule : string -> string -> unit;
  (** [add_rule rl rr] adds a rule [rl => rr] in the environment *)
  get_op_arity : string -> int;
  (** [get_op_arity op] returns the arity of [op] if [op] is a defined operator *)
  is_op : string -> bool;
  (** [is_op op] tells if [op] is defined in the environment *)
  define : string -> Syntax_tree.base_expr -> unit;
  (** [define def e] sets the definition [def] to the value [e] in the current scope *)
  is_defined : string -> bool;
  (** [is_defined def] tells if [def] is defined int the environment*)
  get_def : string -> Syntax_tree.base_expr;
  (** [get_def def] returns the definition associated to [def] if [def] is defined *)
  get_ops : unit -> (string * int) list;
  (** [get_ops ()] returns the list of defined operators with their arity *)
  get_vars :unit -> string list;
  (** [get_vars ()] returns the list of defined variables *)
  get_defs : unit -> (string * Syntax_tree.base_expr) list;
  (** [get_defs ()] returns the list of definitions with their values *)
  get_rules : unit -> (string * string) list;
  (** [get_rules ()] returns the rewriting rules stored by the environment *)
  set_current_dir : string -> unit;
  (** [set_current_dir dir] associate the environment with the directory [dir] *)
  get_current_dir : unit -> string;
  (** [get_current_dir ()] returns the directory associated to the environment *)
  push : (string * Syntax_tree.base_expr) list -> unit;
  (** [push l] defines a new scope for the definitions contained in l

      In this new scope, the definitions take their associated value in l *)
  pop : string list -> unit;
  (** [pop l] retore the previous scope for the definitions contained in l *)
  defun : string -> string list -> Syntax_tree.insts -> unit;
  (** [defun f prms code] defines a new function [f] in the environment,
      
      with parameters [prms] and code [code]*)
  apply_fun : string -> Syntax_tree.expr list -> Syntax_tree.inst;
  (** [apply_fun f vals] returns the code of function [f] where referenced parameters

      have been replaced by the values provided by [vals] *)
  is_fun : string -> bool;
  (** [is_fun f] tells if [f] is a function defined in the environment *)
  get_funs : unit -> (string * (string list * Syntax_tree.insts)) list;
  (** [get_funs ()] returns the list of functions defined in the environment *)
  
  set_lines : (int * int) list -> unit;
  get_line : unit -> int;
  set_stream : char Stream.t -> unit;
  set_stream_line : unit -> unit;
  set_program_name : string -> unit;
  get_program_name : unit -> string;

}

(** [get_new_environment ()] returns an empty environment *)
val get_new_environment : unit -> environment

(** [merge_environments eref e] updates [eref] environment with [e] data*)
val merge_environments : environment -> environment -> unit

val print_debug : environment -> unit

val get_parsed_programs : unit -> (string * environment) list
val set_parsed_program : string -> environment -> unit

val debug_mode : bool ref
