open JClass
open JBasics
open JFile
open Stubs
open Utils

type program = {
  p_structure : interface_or_class list;
  p_max_pcs : int;
  p_max_locals : int
}

let nilprogram : program = {p_structure = []; p_max_pcs = 0; p_max_locals = 0}


(* returns the highest number of local variables of the all the
   methods of the class *)
let max_number_of_locals (ic : interface_or_class) : int =
  match ic with
    | `Class c ->
	let f _ (jm : jmethod) (n : int) =
	  match jm with
	    | AbstractMethod _ -> n
	    | ConcreteMethod cm ->
		(match cm.cm_implementation with
		   | Native ->
		       max (1 + (List.length cm.cm_signature.ms_parameters)) n
		   | Java lazycode ->
		       let code = Lazy.force lazycode in
			 max (code.c_max_locals) n
		)
	in MethodMap.fold f c.c_methods 2
    | _ -> 0

(* associates absolute pc to each opcode in opcodes *)
let mark_opcodes (code : opcodes) : (int * opcode) array =
  let marked_code = Array.make (Array.length code) (-1,OpInvalid)
  and pc = ref 0 in
    Array.iteri (fun i op ->
		 match op with
		   | e when e <> OpInvalid ->
		       marked_code.(i) <- (!pc,op); pc := !pc + 1
		   | _ -> ()) code;
    marked_code

(* replaces relative jumps to absolute ones and put all OpInvalid at the end *)
let to_absolute_pc (code : opcodes) : unit =
  try
    let mcode = mark_opcodes code
    and rwcode = Array.make (Array.length code) OpInvalid
    and cind = ref 0 in
      Array.iteri
	(fun i (_,op) ->
	   (match op with
	      | OpIf (v,n) ->
		  rwcode.(!cind) <- OpIf(v,fst mcode.(i + n));
		  cind := !cind + 1
	      | OpIfCmp (v,n) ->
		  rwcode.(!cind) <- OpIfCmp(v,fst mcode.(i + n));
		  cind := !cind + 1
	      | OpGoto n ->
		  rwcode.(!cind) <- OpGoto(fst mcode.(i + n));
		  cind := !cind + 1
	      | OpJsr n ->
		  rwcode.(!cind) <- OpJsr(fst mcode.(i + n));
		  cind := !cind + 1
	      | OpRet n ->
		  rwcode.(!cind) <- OpRet(fst mcode.(i + n));
		  cind := !cind + 1
	      | e when e <> OpInvalid ->
		  rwcode.(!cind) <- e;
		  cind := !cind + 1
	      | _ -> ())) mcode;
      Array.iteri
	(fun n op -> code.(n) <- op) rwcode;
  with _ -> failwith "Error in opcodes, wrong relative jumps"


(* applies to_absolute_pc to each method of a class *)
let to_absolute_pcs (ic : interface_or_class) : unit =
  match ic with
    | `Class c ->
	(let f _ (jm : jmethod) =
	   match jm with
	     | AbstractMethod _ -> ()
	     | ConcreteMethod cm ->
		 (match cm.cm_implementation with
		    | Native -> ()
		    | Java lazycode ->
			let code = Lazy.force lazycode in
			  to_absolute_pc code.c_code
		 )
	 in MethodMap.iter f c.c_methods)
    | _ -> ()

(* returns the highest instruction index of the all the methods of  the class *)
(* condition : to_absolute_pc have been applied !!! *)
let max_number_of_pcs (ic : interface_or_class) : int =
  match ic with
    | `Class c ->
	let code_size (code : opcodes) =
	  let i = ref 0 in
	    while (!i < Array.length code && code.(!i) <> OpInvalid) do
	      i := !i + 1
	    done;
	    !i
	in let f _ (jm : jmethod) (n : int) =
	    match jm with
	      | AbstractMethod _ -> n
	      | ConcreteMethod cm ->
		  (match cm.cm_implementation with
		     | Native -> n
		     | Java lazycode ->
			 let code = Lazy.force lazycode in
			   max (code_size code.c_code) n
		  )
	in MethodMap.fold f c.c_methods 0
    | _ -> 0


let parse_class (classpath : string) (name : string) =
  if (List.mem name (get_stubs_names())) then
    get_stub name
  else
    let c = get_class (class_path classpath) name in
      to_absolute_pcs c;
      c

(* [get_external_classes ic] get class names referenced in [ic] *)
let get_external_classes (ic : interface_or_class) : class_name list =
  match ic with
    | `Class c ->
	let add_cond e l =
	  if (e <> c.c_name && not(List.mem e l)) then true else false in
	  Array.fold_left
	    (fun l cst ->
	       match cst with
		 | ConstValue(ConstClass(TClass cn)) when add_cond cn l ->
		     cn :: l
		 | ConstField(cn,_,_) when add_cond cn l ->
		     cn :: l
		 | ConstMethod(TClass cn,_,_) when add_cond cn l ->
		     cn :: l
		 | _ -> l)
	    []
	    c.c_consts
    | _ -> []

(* builds a program *)
(* The root class must stay the first of the list *)
let full_parse (rootclass : string) (classpath : string) =
  let rec full_parse_aux (l_todo : class_name list) (l_done : class_name list)
      (l_ic_done : interface_or_class list) (classpath : string) =
    match (disjonction l_todo l_done) with
      | [] -> l_ic_done
      | hd :: tl ->
	  let c = parse_class classpath (String.concat "." hd) in
	  let external_c = get_external_classes c in
	    full_parse_aux ((disjonction external_c l_done) @ tl) (l_done @ [hd])
	      (l_ic_done @ [c]) classpath in
  let l =
    full_parse_aux [Str.split (Str.regexp "\\.") rootclass] [] [] classpath in
    {p_structure = l @ (disjonction (List.map get_stub (get_stubs_names())) l);
     p_max_pcs = List.fold_left (fun m e -> max (max_number_of_pcs e) m) 0 l;
     p_max_locals = List.fold_left (fun m e -> max (max_number_of_locals e) m) 0 l
    }

