open Utils

type insts =
  | Inst of inst * insts * string * int
  | Nop
and inst =
  | Import of string
  | Load of string
  | Genrule of expr * expr
  | Set of string * expr
  | Setn of string * expr * expr
  | Defun of string * string list * insts
  | LetIn of (string * expr) list * insts
  | If of expr * insts * insts
  | For of string * range * insts
  | Expr of expr
and expr = simple_expr * string * int
and simple_expr =
  | BaseExpr of base_expr
  | BaseTerm of string option * expr list
  | BaseConstructor of string * bool
  | BaseVar of string
  | Eval of string
  | Seq of string * bool * expr
  | CSeq of string * expr
  | USeq of string * expr
  | Getn of expr * expr
  | Subsn of expr * expr * expr
  | Add of expr list
  | Op of expr * expr
  | Mem of expr * expr
  | Cons of expr * expr
  | Len of expr
  | Depth of expr
  | Not of expr
  | And of expr list
  | Or of expr list
  | Concat of string * expr
  | FunCall of string * expr list
  | Equal of expr * expr
  | Sup of expr * expr
  | Inf of expr * expr
  | Different of expr * expr
and base_expr =
  | Term of string option * base_expr list
  | Constructor of string
  | Var of string
and range =
  | Range of expr
  | DblRange of expr * expr

let unit = Term(None,[])

let rec string_of_base_expr e =
  match e with
    | Term(a,b) ->
	(match a with
	  | Some n -> n ^ encaps(String.concat ","
				   (List.map string_of_base_expr b))
	  | None -> encaps(String.concat ","
			     (List.map string_of_base_expr b)))
    | Constructor(c) -> c
    | Var(c) -> "var" ^ encaps(c)

let int_of_base_expr e =
  match e with
    | Term(_,_) -> failwith ("Invalid term " ^ string_of_base_expr e)
    | Constructor(c) ->
	(try int_of_string c
	 with
	   | _ -> failwith ("Invalid term " ^ c))
    | Var(_) -> failwith ("Invalid term " ^ string_of_base_expr e)

let keystring_of_expr e =
  match e with
    | BaseExpr s ->
	(match s with
	  | Term (s,_) -> (match s with
			     | Some s -> s^"(...)"
			     | None -> "(...)")
	  | Constructor(s) -> s
	  | Var(s) -> "var("^s^")")
    | BaseTerm (s,_) -> (match s with
			   | Some s -> s^"(...)"
			   | None -> "(...)")
    | BaseConstructor (s,_) -> s
    | BaseVar s -> "var("^s^")"
    | Eval s -> "$"^s
    | Seq (s,b,_) ->
	if b then "seq(var("^s^"),...)"
	else "seq("^s^",...)"
    | CSeq (s,_) -> "cseq("^s^",...)"
    | USeq (s,_) -> "useq("^s^",...)"
    | Getn _ -> "getn(...,...)"
    | Subsn _ -> "subsn(...,...,...)"
    | Add _ -> "add(...)"
    | Op _ -> "op(...,...)"
    | Mem _ -> "mem(...,...)"
    | Cons _ -> "cons(...,...)"
    | Len _ -> "len(...)"
    | Depth _ -> "depth(...)"
    | Not _ -> "not(...)"
    | And _ -> "and(...)"
    | Or _ -> "or(...)"
    | Concat (s,_) -> "concat("^s^",...)"
    | FunCall (s,_) -> "$"^s^"(...)"
    | Equal _ -> "equal(...,...)"
    | Sup _ -> "sup(...,...)"
    | Inf _ -> "inf(...,...)"
    | Different _ -> "different(...,...)"

let keystring_of_inst i =
  match i with
    | Import s -> "import "^s
    | Load s -> "load "^s
    | Genrule _ -> "genrule(...,...)"
    | Set (s,_) -> "set("^s^",...)"
    | Setn (s,_,_) -> "setn("^s^",...,...)"
    | Defun (s,p,_) -> "defun "^s^"("^(String.concat "," p)^") = (...)"
    | LetIn _ -> "let ... in ..."
    | If _ -> "if (...) then ..."
    | For (s,_,_) -> "for "^s^" ... do(...)"
    | Expr (e,_,_) -> keystring_of_expr e

let term_getn e n =
  match e with
    | Constructor(c) -> failwith ("Invalid index " ^ (string_of_int n)
				  ^ " for term " ^ c)
    | Var(c) -> failwith ("Invalid index " ^ (string_of_int n)
			  ^ " for term " ^ "var" ^ encaps(c))
    | Term(_,b) ->
	(try List.nth b (n - 1)
	 with _ -> failwith ("Invalid index " ^ (string_of_int n)
			     ^ " for term "
			     ^ (string_of_base_expr e)))

let term_setn e n v =
  match e with
    | Constructor(c) -> failwith ("Invalid index " ^ (string_of_int n)
				  ^ " for term " ^ c)
    | Var(c) -> failwith ("Invalid index " ^ (string_of_int n)
			  ^ " for term " ^ "var" ^ encaps(c))
    | Term(a,b) ->
	(try let len = List.length b in
	   assert (n >= 1 && n <= len);
	   let l = ref []
	   and j = ref len in
	     while (!j >= 1) do
	       l := !j :: !l;
	       j := !j - 1
	     done;
	     let args = List.map2 (fun x y -> if (x = n) then v else y) !l b in
	       Term(a,args)
	 with _ -> failwith ("Invalid index " ^ (string_of_int n)
			     ^ " for term "
			     ^ (string_of_base_expr (Term(a,b)))))

let term_add a b =
  match (a,b) with
    | (Constructor(x),Constructor(y)) ->
	(try
	   let sum = (int_of_string x) + (int_of_string y) in
	     Constructor(string_of_int sum)
	 with _ -> failwith ("Invalid operands for add : " ^ x ^ " or " ^ y))
    | _ -> failwith ("Invalid operands for add")

let rec term_symboldepth r e =
  match e with
    | Constructor(_) -> 0
    | Term(Some(a),b) when a = r ->
	1 + List.fold_left (fun x y -> max x (term_symboldepth r y)) 0 b
    | _ -> 0 

let term_and e1 e2 =
  match (e1,e2) with
    | (Constructor(v), _) when v = val_false -> Constructor(val_false)
    | (_, Constructor(v)) when v = val_false -> Constructor(val_false)
    | _ -> Constructor(val_true)

let term_or e1 e2 =
  match (e1,e2) with
    | (Constructor(u), Constructor(v)) when (u, v) =
	(val_false, val_false) -> Constructor(val_false)
    | _ -> Constructor(val_true)
